from typing import Any, Dict, Callable, List, Tuple, AsyncGenerator

from ..types import Update, Message, Response
from ..helpers import ApiHelper

try:
    from starlette.responses import PlainTextResponse
    from starlette.requests import Request
except: pass

import asyncio


class BaseBot:
    def __init__(self, token: str) -> None:
        self.token = token

        self._api = ApiHelper(token)

        self._update_offset = 0

        self._key_text_msgs: Dict[str, Callable[[Message], Any]] = {}
        self._key_cmd_msgs: Dict[str, Callable[[Message], Any]] = {}
        self._file_id_sticker_msgs: Dict[str, Callable[[Message], Any]] = {}
        self._global_msg_types: Dict[str, Callable[[Message], Any]] = {}

        self._func_handlers: List[Tuple[ Callable[[Message], bool], Callable[[Message], Any] ]] = []
        

    async def _process_message(self, message: Message):
        if 'all' in self._global_msg_types:
            asyncio.create_task( self._global_msg_types['all'](message) )

        if message.text:
            if 'text' in self._global_msg_types:
                asyncio.create_task( self._global_msg_types['text'](message) )

            if message.text in self._key_text_msgs:
                asyncio.create_task( self._key_text_msgs[message.text](message) )

            if message.text[1:] in self._key_cmd_msgs:
                asyncio.create_task( self._key_cmd_msgs[message.text[1:]](message) )

        if message.sticker:
            if 'sticker' in self._global_msg_types:
                asyncio.create_task( self._global_msg_types['sticker'](message) )
            
            if (message.sticker.file_unique_id in self._file_id_sticker_msgs):
                asyncio.create_task( self._file_id_sticker_msgs[message.sticker.file_unique_id](message) )

        if message.document:
            if 'document' in self._global_msg_types:
                asyncio.create_task( self._global_msg_types['document'](message) )

        if message.photo:
            if 'photo' in self._global_msg_types:
                asyncio.create_task( self._global_msg_types['photo'](message) )

        for item in self._func_handlers:
            if item[0](message):
                asyncio.create_task( item[1](message) )


    def _process_response(self, json_data: dict, cls):
        response = Response(json_data)
        
        if response.ok:
            return response.result_instance(cls)
        else:
            print(f'Error: {response.error_code} => {response.description}')

    
    async def _process_updates_json(self, json_data: dict):
        updates: List[Update] = self._process_response(json_data, Update)

        for update in updates:
            self._update_offset = update.update_id + 1
            
            if update.message:
                await self._process_message(update.message)
            else:
                assert False, f'Update attribute {update.dict()} not support'
        

    async def polling(self):
        async with self._api.session:
            while True:
                json_data = await self._api.get_updates(timeout=60, offset=self._update_offset)
                await self._process_updates_json(json_data)


    async def __call__(self, scope, receive, send):
        if scope['type'] != 'http': return

        if scope['method'] != 'POST':
            await PlainTextResponse('Method not allowed', 405)(scope, receive, send)

        if (scope['path']) != f'/{self.token}':
            await PlainTextResponse("Unknown path", 404)(scope, receive, send)

        json_data = await Request(scope, receive).json()
        await self._process_updates_json(json_data)

        await PlainTextResponse('success', 200)        


from typing import Any, Optional


class Response:
    ok: bool = None
    result: Optional[Any] = None
    error_code: Optional[int] = None
    description: Optional[str] = None


    def __init__(self, json_data: dict):
        for k, v in json_data.items():
            self.__dict__[k] = v

    def result_instance(self, cls):
        if type(self.result) is list:
            result = [cls(obj) for obj in self.result]
        else:
            result = cls(self.result)

        return result


from typing import Callable, Optional, Any, List

from ..helpers import generate_payload
from .base import BaseBot
from ..types import (
    Message, 
    File,
    ReplyBase,
)
from ..types import ReplyBase

import asyncio


class Bot(BaseBot):
    async def send_msg(
        self,
        chat_id                     : int,
        text                        : str,
        parse_mode                  : Optional[str] = None, 
        disable_web_page_preview    : Optional[bool] = None,
        disable_notification        : Optional[bool] = None,
        reply_to_message_id         : Optional[int] = None,
        allow_sending_without_reply : Optional[int] = None,
        reply_markup                : Optional[ReplyBase] = None
    ) -> Message:
        data = generate_payload(locals().copy())
        response = await self._api.make_request('sendMessage', data)
        return self._process_response(response, Message)


    async def send_sticker(
        self,
        chat_id                     : int,
        sticker                     : str,
        disable_notification        : Optional[bool] = None,
        reply_to_message_id         : Optional[int] = None,
        allow_sending_without_reply : Optional[int] = None,
        reply_markup                : Optional[ReplyBase] = None
    ):
        data = generate_payload(locals().copy())
        response = await self._api.make_request('sendSticker', data)
        return self._process_response(response, Message)


    async def download_file(self, file_id: str, save_path: str=None):
        file: File = await self.get_file(file_id)
        return await self._api.download_file(file.file_path, save_path)


    async def get_file(self, file_id: str):
        data = { 'file_id': file_id }
        response = await self._api.make_request('getFile', data=data)
        return self._process_response(response, File)


    def on_message(
        self, *,
        texts   : List[str] = None,
        cmds    : List[str] = None,
        stickers: List[str] = None,
        msg_type: str = None,
        func    : Callable[[Message], bool] = None
    ):
        def decorator(d_func: Callable[[Message], Any]):
            if texts:
                self._key_text_msgs.update( {text: d_func for text in texts} )

            if cmds:
                self._key_cmd_msgs.update( {cmd: d_func for cmd in cmds} )

            if stickers:
                self._file_id_sticker_msgs.update( {sticker: d_func for sticker in stickers} )
            
            if msg_type:
                self._global_msg_types[msg_type] = d_func

            if func:
                self._func_handlers.append( (func, d_func) )

        return decorator

            
    def run(self):
        try:
            asyncio.run(self.polling())
        except KeyboardInterrupt:
            quit()



from typing import List, Optional

from . import ReplyBase
from .. import TeleObj, Field
from ...helpers import generate_payload


class KeyboardButton(TeleObj):
    text: str = Field()
    request_contact: Optional[bool] = Field()
    request_location: Optional[bool] = Field()

    def __init__(self, 
        text: str, 
        request_contact: Optional[bool]=None, 
        request_location: Optional[bool]=None
    ):
        payload = generate_payload(locals().copy())
        super().__init__(payload)


class ReplyKeyboardMarkup(ReplyBase):
    keyboard: List[List[KeyboardButton]] = Field(KeyboardButton, two_dim=True)
    resize_keyboard: Optional[bool] = Field()
    one_time_keyboard: Optional[bool] = Field()
    selective: Optional[bool] = Field()

    def __init__(
        self, 
        keyboard: List[List[KeyboardButton]],
        resize_keyboard: Optional[bool] = None, 
        one_time_keyboard: Optional[bool] = None,
        selective: Optional[bool] = None
    ):
        payload = generate_payload(locals().copy())
        super().__init__(payload)



from .field import Field
from .base import TeleObj

from .login_url import LoginUrl
from .chat import Chat
from .user import User
from .document import Document
from .photo_size import PhotoSize
from .message_entity import MessageEntity
from .mask_position import MaskPosition
from .sticker import Sticker
from .message import Message
from .file import File

from .update import Update
from .response import Response

from .keyboard import (
    ReplyBase,
    ReplyKeyboardMarkup, KeyboardButton,
    InlineKeyboardMarkup, InlineKeyboardButton
)


from typing import Optional
from . import TeleObj, Field


class PhotoSize(TeleObj):
    file_id: str = Field()
    file_unique_id: str = Field()
    width: int = Field()
    height: int = Field()
    file_size: Optional[int] = Field()


from typing import Optional
from . import TeleObj, Field

from .photo_size import PhotoSize


class Document(TeleObj):
    file_id: str = Field()
    file_unique_id: str = Field()
    thumb: Optional[PhotoSize] = Field(PhotoSize)
    file_name: Optional[str] = Field()
    mime_type: Optional[str] = Field()
    file_size: Optional[int] = Field()


from . import Field


class TeleObj:
    _corrects = {}

    def __init__(self, json_data: dict):
        cur_type: TeleObj = type(self)

        for key, value in json_data.items():
            field: Field = cur_type.__dict__.get(key)

            if field is None:
                key = cur_type._corrects.get(key)
                field = cur_type.__dict__.get(key)

            if field is None:
                #raise TypeError(f'{type(self).__name__} got an unexpected keyword argument \'{key}\'')
                continue

            self.__dict__[key] = field.make_obj(value, self)

    
    def dict(self):
        corrects = { field_name: json_name for json_name, field_name in self._corrects.items() }

        res = {}

        for key, value in self.__dict__.items():
            if key in corrects:
                key = corrects[key]

            if type(value) is list:
                res[key] = [ 
                    item.dict() if isinstance(item, TeleObj) 
                    else [obj.dict() if isinstance(obj, TeleObj) else obj for obj in item] if isinstance(item, list)
                    else item
                    
                    for item in value 
                ]
            elif isinstance(value, TeleObj):
                res[key] = value.dict()
            else:
                res[key] = value

        return res


from typing import Optional

from . import (
    TeleObj, 
    Field,
    PhotoSize,
    MaskPosition
)


class Sticker(TeleObj):
    file_id         : str = Field()
    file_unique_id  : str = Field()
    width           : int = Field()
    height          : int = Field()
    is_animated     : bool = Field()
    thumb           : Optional[PhotoSize] = Field(PhotoSize)
    emoji           : Optional[str] = Field()
    set_name        : Optional[str] = Field()
    mask_position   : Optional[MaskPosition] = Field(MaskPosition)
    file_size       : Optional[int] = Field()


from typing import List, Optional

from .types import TeleObj

import aiohttp


def generate_payload(args: dict):
    del args['self']
    data = { 
        k: v.dict() if isinstance(v, TeleObj) else v
        for k, v in args.items() if v is not None 
    }

    return data

class ApiHelper:
    def __init__(self, token):
        self.token = token

        self._session: aiohttp.ClientSession = None

        self.request_config = { 'ssl': False }

    @property
    def session(self):
        if self._session is None or self._session.closed:
            self._session = aiohttp.ClientSession()
        return self._session

    async def make_request(self, method: str, data: dict = None) -> Optional[dict]:
        url = f'https://api.telegram.org/bot{self.token}/{method}'
        async with self.session.post(url, json=data, **self.request_config) as resp:
            return await resp.json()

    async def download_file(self, file_path: str, save_path: str=None):
        async with self.session.get(f'https://api.telegram.org/file/bot{self.token}/{file_path}', **self.request_config) as resp:
            if save_path:
                with open(save_path, 'wb') as file:
                    data = await resp.content.read()
                    file.write(data)
            else:
                return await resp.content.read()

    async def get_updates(
            self,
            offset: Optional[int]=None, 
            limit: Optional[int]=None, 
            timeout: Optional[int]=None,
            allowed_updates: Optional[List[str]] = None
        ) -> Optional[dict]:
        data = generate_payload(locals().copy())

        return await self.make_request('getUpdates', data=data)


class Field:
    def __init__(self, cls=None, *, parent=None, two_dim=False):
        self.cls = cls
        self.parent = parent
        self.two_dim = two_dim

    def __get__(self, instance, cls): pass

    def make_obj(self, value, instance):
        if self.parent is list:
            if self.two_dim:
                return [ [self.cls(obj) for obj in item_list] for item_list in value ]
            else:
                return [self.cls(obj) for obj in value] if self.cls else value

        elif self.cls is ...:
            return type(instance)(value)

        elif self.cls is not None:
            return self.cls(value)

        else:
            return value


from typing import Optional
from . import TeleObj, Field


class User(TeleObj):
    id: int = Field()
    is_bot: bool = Field()
    first_name: str = Field()
    last_name: Optional[str] = Field()
    username: Optional[str] = Field()
    language_code: Optional[str] = Field()
    can_join_groups: Optional[bool] = Field()
    can_read_all_group_messages: Optional[bool] = Field()
    supports_inline_queries: Optional[bool] = Field()



from typing import Optional, List

from .. import Field, TeleObj, LoginUrl
from ...helpers import generate_payload

from . import ReplyBase


class InlineKeyboardButton(TeleObj):
    text                            : str = Field()
    url                             : Optional[str] = Field()
    login_url                       : Optional[LoginUrl] = Field(LoginUrl)
    callback_data                   : Optional[str] = Field()
    switch_inline_query             : Optional[str] = Field()
    switch_inline_query_current_chat: Optional[str] = Field()
    pay                             : Optional[bool] = Field()

    def __init__(
        self,
        text                            : str,
        url                             : Optional[str] = None,
        login_url                       : Optional[LoginUrl] = None,
        callback_data                   : Optional[str] = None,
        switch_inline_query             : Optional[str] = None,
        switch_inline_query_current_chat: Optional[str] = None,
        pay                             : Optional[bool] = None,
    ):
        data = generate_payload(locals().copy())
        super().__init__(data)


class InlineKeyboardMarkup(ReplyBase):
    inline_keyboard: List[List[InlineKeyboardButton]] = Field(InlineKeyboardButton, two_dim=True)

    def __init__(
        self,
        inline_keyboard: List[List[InlineKeyboardButton]]
    ):
        data = generate_payload(locals().copy())
        super().__init__(data)


from typing import Optional
from . import TeleObj, Field

from .message import Message


class Update(TeleObj):
    update_id: int = Field()
    message: Optional[Message] = Field(Message)
    edited_message: Optional[Message] = Field(Message)
    channel_post: Optional[Message] = Field(Message)
    edited_channel_post: Optional[Message] = Field(Message)
    


from typing import Optional
from . import TeleObj, Field

from .user import User


class MessageEntity(TeleObj):
    type: str = Field()
    offset: int = Field()
    length: int = Field()
    url: Optional[str] = Field()
    user: Optional[User] = Field(User)
    language: Optional[str] = Field()


import setuptools
import re

with open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()

with open("teleNex/__init__.py", 'r', encoding="utf-8") as f:
    version = re.search(r'__version__ = "(.*?)"', f.read()).group(1)

setuptools.setup(
    name                          = 'TeleNex',
    version                       = version,
    author                        = 'iNex',
    author_email                  = 'inex552@gmail.com',
    description                   = 'TeleNex - это Frimework для простого создания асинхронных Telegram ботов',
    long_description              = long_description,
    long_description_content_type = "text/markdown",
    url                           = 'https://github.com/inex550/TeleNex',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages(),
    install_requires=[
        'starlette>=0.14.2',
        'aiohttp>=3.7.4.post0'
    ],
    python_requires='>=3.7'
)


from typing import Optional, List

from . import TeleObj, Field
from . import (
    User, 
    Chat, 
    MessageEntity, 
    Document, 
    PhotoSize, 
    Sticker
)


class Message(TeleObj):
    _corrects = {
        'from': 'from_user'
    }

    message_id: int = Field()
    from_user: Optional[User] = Field(User)
    sender_chat: Optional[Chat] = Field(Chat)
    date: int = Field()
    chat: Chat = Field(Chat)
    forward_from: Optional[User] = Field(User)
    forward_from_chat: Optional[Chat] = Field(Chat)
    forward_from_message_id: Optional[int] = Field()
    forward_signature: Optional[str] = Field()
    forward_sender_name: Optional[str] = Field()
    forward_date: Optional[int] = Field()
    reply_to_message: Optional['Message'] = Field(...)
    via_bot: Optional[User] = Field(User)
    edit_date: Optional[int] = Field()
    media_group_id: Optional[str] = Field()
    author_signature: Optional[str] = Field()
    text: Optional[str] = Field()
    entities: Optional[List[MessageEntity]] = Field(MessageEntity, parent=list)
    document: Optional[Document] = Field(Document)
    photo: Optional[List[PhotoSize]] = Field(PhotoSize, parent=list)
    sticker: Optional[Sticker] = Field(Sticker)


from typing import Optional
from . import TeleObj, Field


class Chat(TeleObj):
    id: int = Field()
    type: str = Field()
    title: Optional[str] = Field()
    username: Optional[str] = Field()
    first_name: Optional[str] = Field()
    last_name: Optional[str] = Field()
    bio: Optional[str] = Field()
    description: Optional[str] = Field()
    invite_link: Optional[str] = Field()


from typing import Optional
from . import TeleObj, Field


class File(TeleObj):
    file_id: str = Field()
    file_unique_id: str = Field()
    file_size: Optional[str] = Field()
    file_path: Optional[str] = Field()